#
# Licensed Materials - Property of IBM
# 5655-FLW (C) Copyright IBM Corporation 2008, 2010. All Rights Reserved.
# US Government Users Restricted Rights- Use, duplication or disclosure
# restricted by GSA ADP Schedule Contract with IBM Corp.
#

import os.path, sys

from java.lang import Integer, System, Thread
from java.util import HashMap
from java.net import URL, URLClassLoader
# only classes from plugins/*runtime*.jar and plugins/*admin*.jar are visible to Jython in WAS 7
from com.ibm.bpe.database.migration import DataMigrationMgr

execfile(os.path.join(System.getProperty("was.install.root"), "ProcessChoreographer/config/bpclib.py"))
__file__ = "migrateDB.py"

#*******************************************************************************************************
# globals
#*******************************************************************************************************
MIN_SLICE = 10
MAX_SLICE = 50000
DEFAULT_SLICE = 500

#*******************************************************************************************************
# print usage and error message, if any
#*******************************************************************************************************
def usage(errmsg = None):
  if errmsg:
    print
    print "ERROR: " + errmsg
  #endif
  print
  print "Usage: wsadmin -conntype NONE -f " + __file__
  print "       (-server <serverName> | -cluster <clusterName>)"
  print "       [-dbUser <userID>] -dbPassword <password>"
  print "       [-dbSchema <schema>]"
  print "       [-slice <slice>]"
  print "       [-skipTablespaceMigration TRUE]"
  print
  print "Migrates the data in a pre-6.2 Business Process Choreographer database. This script must"
  print "be run in disconnected mode, that is, the application server or cluster must not be running."
  print
  print "Note that this script must be run on the node of the server (when '-server' was specified) or"
  print "on the node of a cluster member (when '-cluster' was specified). That is, it cannot be run on"
  print "the deployment manager node in an ND environment."
  print
  print "The -server and -cluster parameters denote the Business Process Choreographer"
  print "configuration for which to migrate the database."
  print "The -dbUser and -dbPassword parameters are used to authenticate with the database."
  print "The optional -dbSchema parameter can be used to overwrite an implicit schema qualifier."
  print "The optional -slice parameter can be used to overwrite the default transaction slice size (" + str(DEFAULT_SLICE) + ")."
  print "Its value must be between " + str(MIN_SLICE) + " and " + str(MAX_SLICE) + "."
  print "The optional -skipTablespaceMigration parameter allows to skip the tablespace migration for DB2"
  print "if the tables are already contained in a suitable table space."
#enddef


#*******************************************************************************************************
# validate integer parameter
#*******************************************************************************************************
def validateIntValue(props, key, minValue, maxValue):
  errorMsg = None
  try:
    props[key] = int(props[key])
  except ValueError:
    errorMsg = "Error: '" + key + "' must be an integer."
  else:
    if props[key] < minValue or props[key] > maxValue:
      errorMsg = "Error: '" + key + "' must be in the range from " + str(minValue) + " to " + str(maxValue) + "."
    #endif
  #endtry
  return errorMsg
#enddef

#*******************************************************************************************************
# main
#*******************************************************************************************************

# ----------------------------------------------------
# make sure the script is running in disconnected mode
if wsadmin.getConnectionStatus() == WsAdmin.CONNECTED:
  usage(__file__ + " cannot be run in connected mode, re-run using the '-conntype NONE' option.")

if wsadmin.getConnectionStatus() != WsAdmin.DISCONNECTED: # connected or error -> exit
  sys.exit(1)

# ----------------------
# parse the command line
props = { "server": None, "cluster": None, "slice": DEFAULT_SLICE, "dbUser": None, "dbPassword": None, "force": "FALSE", "skipTablespaceMigration": "FALSE" }
props = wsadmin.parse(sys.argv, props)
props["node"] = System.getProperty("local.node")

# -----------------------------
# check command line parameters
if not props["server"] and not props["cluster"]:
  usage("One of the '-server' and '-cluster' parameters must be specified.")
  sys.exit(1)
#endif

# ------------------------------------------------------------
# validate the 'slice' value
sliceError = validateIntValue(props, "slice", MIN_SLICE, MAX_SLICE)
if sliceError:
  usage(sliceError)
  sys.exit(0)
#endif

# -----------------------------------------------
# check if the scope is a valid deployment target
scope = Scope(nodeName = props["node"], serverName = props["server"], clusterName = props["cluster"])
if scope.convertToCluster():
  print "Warning: '" + props["server"] + "' is a member of cluster '" + scope.clusterName + "', adjusting scope to cluster."
#endif

if not scope.isDeploymentTarget():
  usage(str(scope) + " is not a valid deployment target.")
  sys.exit(1)
#endif

if scope.getCluster():
  foundMemberOnLocalNode = 0
  for member in wsadminToList(scope.list("ClusterMember")):
    if AdminConfig.showAttribute(member, "nodeName") == props["node"]:
      foundMemberOnLocalNode = 1
      break
    #endif
  #endfor
  if not foundMemberOnLocalNode:
    usage("No member of cluster " + props["cluster"] + " is defined on local node " + props["node"] + ".")
    sys.exit(1)
  #endif
#endif

# ------------------------------------------------------------
# find the BPC data source and query its provider's properties
dataSource = scope.lookup(findDataSource, "jdbc/BPEDB")

if not dataSource:
  usage("Data source 'jdbc/BPEDB' not found in scope " + str(scope))
  sys.exit(0)
#endif

provider = AdminConfig.showAttribute(dataSource, "provider")
providerType = AdminConfig.showAttribute(provider, "providerType")
classpath = AdminConfig.showAttribute(provider, "classpath")
className = AdminConfig.showAttribute(provider, "implementationClassName")
authAlias = AdminConfig.showAttribute(dataSource, "authDataAlias")

if providerType != "Cloudscape JDBC Provider (XA)" and providerType != "Derby JDBC Provider (XA)" and providerType != "Derby JDBC Provider 40 (XA)" and not props["dbPassword"]:
  usage("The '-dbPassword' parameter is required for " + providerType + ".")
  sys.exit(0)
#endif

authUser = None
if authAlias and authAlias != "":
  security = AdminConfig.getid(scope.getCell() + "Security:/")
  authDataEntries = wsadminToList(AdminConfig.showAttribute(security, "authDataEntries"))
  for authDataEntry in authDataEntries:
    if AdminConfig.showAttribute(authDataEntry, "alias") == authAlias:
      authUser = AdminConfig.showAttribute(authDataEntry, "userId")
      break
    #endif
  #endfor
#endif

# ---------------------------------
# compute the effective schema name
if not props["dbUser"] and authUser and authUser != "":
  props["dbUser"] = authUser
#endif

if props.has_key("dbSchema") and props["dbSchema"]:
  schema = props["dbSchema"]
  print "Using supplied schema qualifier '" + schema + "'."
else:
  dataStore = scope.list("BPCDataStore")
  # WASX7017E: Exception received while running file "migrateDB.py"; exception information: com.ibm.ws.scripting.ScriptingException:
  # WASX7444E: Invalid parameter value "" for parameter "config id" on command "showAttribute"
  # @EF todo -> write wsadmin.showAttribute(). How: static methods?
  if dataStore and dataStore != "":
    schema = AdminConfig.showAttribute(dataStore, "schemaName")
  else:
    print "Warning: BPCDataStore not found in " + str(scope) + ", assuming no schema qualifier has been configured."
    print "This can happen if migrating from WPS 6.0.2 without running the BPMMigrateCluster script."
    schema = None
  #endif
  if schema and schema != "":
    print "Using configured schema qualifier '" + schema + "'."
  elif authUser and authUser != "":
    schema = authUser
    print "Using implicit schema qualifier '" + schema + "'."
  #endif
#endif

# --------------------------------------------------------
# prepare a java.util.Map of data source custom properties
resourceProperties = AdminConfig.showAttribute(AdminConfig.showAttribute(dataSource, "propertySet"), "resourceProperties")
properties = HashMap()
url = None
for property in wsadminToList(resourceProperties):
  name = AdminConfig.showAttribute(property, "name")
  value = AdminConfig.showAttribute(property, "value")
  if (providerType.startswith("Derby") or providerType.startswith("Cloudscape")) and name == "databaseName":
    # the migrated provider still has providerType="Cloudscape JDBC Provider (XA)" although it is based on Derby now
    value = os.path.normpath(AdminTask.expandBPCVariables(scope.asParameterList(["-variableString", value], 1)))
  #endif
  if providerType.startswith("Oracle") and name == "URL":
    pattern = re.compile(r"jdbc:oracle:(?P<driverType>[^:]+):@(?:(?P<serverName>[^:]+):)?(?:(?P<portNumber>[^:]+):)?(?P<databaseName>.*)")
    url = pattern.match(value)
  #endif
  properties.put(name, value)
#endfor

# CommonDB only sets URL, fill in driverType, serverName, portNumber, and databaseName if these are not set
if providerType.startswith("Oracle") and url and url.group("driverType") and not properties["driverType"]:
  properties.put("driverType", url.group("driverType"))
  print "Set driverType=" + str(url.group("driverType")) + " from URL=" + str(url.group(0)) + "."
  if url.group("serverName"):
    properties.put("serverName", url.group("serverName"))
    print "Set serverName=" + str(url.group("serverName")) + " from URL=" + str(url.group(0)) + "."
    if url.group("portNumber"):
      properties.put("portNumber", url.group("portNumber"))
      print "Set portNumber=" + str(url.group("portNumber")) + " from URL=" + str(url.group(0)) + "."
    #endif
  #endif
  if url.group("databaseName"):
    properties.put("databaseName", url.group("databaseName"))
    print "Set databaseName=" + str(url.group("databaseName")) + " from URL=" + str(url.group(0)) + "."
  #endif
#endif

# -------------------------------------------------------------
# set up classpath for JDBC provider, based on settings in WCCM
classpath = AdminTask.expandBPCVariables(scope.asParameterList(["-variableString", classpath], 1))
urls = []
for path in classpath.split(";"):
  path = os.path.normpath(path)
  if not os.path.exists(path):
    print "File " + path + " does not exist, ignoring."
  else:
    url = URL("file", "", -1, path)
    print "File " + path + " exists, adding it to the classpath."
    urls.append(url)
  #endif
#endfor

classloader = URLClassLoader(urls, Thread.currentThread().getContextClassLoader())
Thread.currentThread().setContextClassLoader(classloader)

# -------------------------------------------------------------
# set options for data migration
options= HashMap()
options.put(DataMigrationMgr.SLICE_SIZE, Integer(props["slice"]))
options.put(DataMigrationMgr.FORCE, props["force"])
options.put(DataMigrationMgr.SKIP_TABLESPACE, props["skipTablespaceMigration"])

# ----------------------------
# invoke the DB migration tool
DataMigrationMgr.start(props["dbUser"], props["dbPassword"], schema, className, properties, options)

# ------------------------------------------------------------------------------
# check if custom tables have been configured and if so, print a warning message
# (see also http://www-1.ibm.com/support/docview.wss?uid=swg27010849&aid=1)
bfm = scope.list("BPELFlowManager")
if bfm == "":
  print "ERROR: The Process Choreographer configuration is corrupted - no 'BPELFlowManager' found in " + str(scope) + "."
#endif
props = wsadminToList(AdminConfig.showAttribute(bfm, "properties"))
for prop in props:
  if AdminConfig.showAttribute(prop, "name") == "customTableDefinition":
    value = AdminConfig.showAttribute(prop, "value")
    print "*******************************************************************************"
    print "NOTE: Custom property 'customTableDefinition' has been set to"
    print "  " + value
    value = os.path.normpath(AdminTask.expandBPCVariables(scope.asParameterList(["-variableString", value], 1)))
    if not os.path.isfile(value):
      print "which resolves to"
      print "  " + value
      print "but this file doesn't exists locally."
    #endif
    print "If you were using a custom table definition file for named materialized views,"
    print "they have been dropped by this data migration script. WebSphere Process Server"
    print "will only be able to recreate the materialized views during server startup"
    print "if the 'customTableDefinition' property points to a custom table definition"
    print "file that is accessible by WebSphere Process Server."
    print "If you were using a custom table definition file for custom tables, they must"
    print "be dropped and recreated manually now."
    print "*******************************************************************************"
    break
  #endif
#endfor

print __file__ + " finished."
